package java.util.concurrent;

import java.util.ConcurrentModificationException;
import java.util.AbstractList;
import java.util.Objects;
import java.util.NoSuchElementException;
import java.util.Spliterators;
import java.util.Spliterator;
import java.util.ListIterator;
import sun.misc.SharedSecrets;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.Comparator;
import java.util.function.UnaryOperator;
import java.util.function.Predicate;
import java.util.function.Consumer;
import java.util.Iterator;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Collection;
import sun.misc.Unsafe;
import java.util.concurrent.locks.ReentrantLock;
import java.io.Serializable;
import java.util.RandomAccess;
import java.util.List;

public class CopyOnWriteArrayList<E> implements List<E>, RandomAccess, Cloneable, Serializable
{
    private static final long serialVersionUID = 8673264195747942595L;
    final transient ReentrantLock lock;
    private transient volatile Object[] array;
    private static final Unsafe UNSAFE;
    private static final long lockOffset;
    
    final Object[] getArray() {
        return this.array;
    }
    
    final void setArray(Object[] array) {
        this.array = array;
    }
    
    public CopyOnWriteArrayList() {
        this.lock = new ReentrantLock();
        this.setArray(new Object[0]);
    }
    
    public CopyOnWriteArrayList(Collection<? extends E> collection) {
        this.lock = new ReentrantLock();
        Object[] array;
        if (collection.getClass() == CopyOnWriteArrayList.class) {
            array = ((CopyOnWriteArrayList)collection).getArray();
        }
        else {
            array = collection.toArray();
            if (collection.getClass() != ArrayList.class) {
                array = Arrays.copyOf(array, array.length, Object[].class);
            }
        }
        this.setArray(array);
    }
    
    public CopyOnWriteArrayList(E[] original) {
        this.lock = new ReentrantLock();
        this.setArray(Arrays.copyOf(original, original.length, Object[].class));
    }
    
    @Override
    public int size() {
        return this.getArray().length;
    }
    
    @Override
    public boolean isEmpty() {
        return this.size() == 0;
    }
    
    private static boolean eq(Object o, Object obj) {
        return (o == null) ? (obj == null) : o.equals(obj);
    }
    
    private static int indexOf(Object o, Object[] array, int i, int n) {
        if (o == null) {
            while (i < n) {
                if (array[i] == null) {
                    return i;
                }
                ++i;
            }
        }
        else {
            while (i < n) {
                if (o.equals(array[i])) {
                    return i;
                }
                ++i;
            }
        }
        return -1;
    }
    
    private static int lastIndexOf(Object o, Object[] array, int i) {
        if (o == null) {
            while (i >= 0) {
                if (array[i] == null) {
                    return i;
                }
                --i;
            }
        }
        else {
            while (i >= 0) {
                if (o.equals(array[i])) {
                    return i;
                }
                --i;
            }
        }
        return -1;
    }
    
    @Override
    public boolean contains(Object o) {
        Object[] array = this.getArray();
        return indexOf(o, array, 0, array.length) >= 0;
    }
    
    @Override
    public int indexOf(Object o) {
        Object[] array = this.getArray();
        return indexOf(o, array, 0, array.length);
    }
    
    public int indexOf(E e, int n) {
        Object[] array = this.getArray();
        return indexOf(e, array, n, array.length);
    }
    
    @Override
    public int lastIndexOf(Object o) {
        Object[] array = this.getArray();
        return lastIndexOf(o, array, array.length - 1);
    }
    
    public int lastIndexOf(E e, int n) {
        return lastIndexOf(e, this.getArray(), n);
    }
    
    public Object clone() {
        try {
            CopyOnWriteArrayList<Object> list = (CopyOnWriteArrayList)super.clone();
            list.resetLock();
            return list;
        }
        catch (CloneNotSupportedException ex) {
            throw new InternalError();
        }
    }
    
    @Override
    public Object[] toArray() {
        Object[] array = this.getArray();
        return Arrays.copyOf(array, array.length);
    }
    
    @Override
    public <T> T[] toArray(T[] array) {
        Object[] array2 = this.getArray();
        int length = array2.length;
        if (array.length < length) {
            return (T[])Arrays.copyOf(array2, length, array.getClass());
        }
        System.arraycopy(array2, 0, array, 0, length);
        if (array.length > length) {
            array[length] = null;
            return array;
        }
        return array;
    }
    
    private E get(Object[] array, int n) {
        return array[n];
    }
    
    @Override
    public E get(int n) {
        return this.get(this.getArray(), n);
    }
    
    @Override
    public E set(int n, E e) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = this.getArray();
            Object value = this.get(array, n);
            if (value != e) {
                Object[] copy = Arrays.copyOf(array, array.length);
                copy[n] = e;
                this.setArray(copy);
            }
            else {
                this.setArray(array);
            }
            return value;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean add(E e) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = this.getArray();
            int length = array.length;
            Object[] copy = Arrays.copyOf(array, length + 1);
            copy[length] = e;
            this.setArray(copy);
            return true;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public void add(int i, E e) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = this.getArray();
            int length = array.length;
            if (i > length || i < 0) {
                throw new IndexOutOfBoundsException("Index: " + i + ", Size: " + length);
            }
            int n = length - i;
            Object[] copy;
            if (n == 0) {
                copy = Arrays.copyOf(array, length + 1);
            }
            else {
                copy = new Object[length + 1];
                System.arraycopy(array, 0, copy, 0, i);
                System.arraycopy(array, i, copy, i + 1, n);
            }
            copy[i] = e;
            this.setArray(copy);
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public E remove(int n) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = this.getArray();
            int length = array.length;
            Object value = this.get(array, n);
            int n2 = length - n - 1;
            if (n2 == 0) {
                this.setArray(Arrays.copyOf(array, length - 1));
            }
            else {
                Object[] array2 = new Object[length - 1];
                System.arraycopy(array, 0, array2, 0, n);
                System.arraycopy(array, n + 1, array2, n, n2);
                this.setArray(array2);
            }
            return value;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean remove(Object o) {
        Object[] array = this.getArray();
        int index = indexOf(o, array, 0, array.length);
        return index >= 0 && this.remove(o, array, index);
    }
    
    private boolean remove(Object o, Object[] array, int index) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array2 = this.getArray();
            int length = array2.length;
            Label_0135: {
                if (array != array2) {
                    int min = Math.min(index, length);
                    int i = 0;
                    while (i < min) {
                        if (array2[i] != array[i] && eq(o, array2[i])) {
                            index = i;
                            break Label_0135;
                        }
                        ++i;
                    }
                    if (index >= length) {
                        return false;
                    }
                    if (array2[index] != o) {
                        index = indexOf(o, array2, index, length);
                        if (index < 0) {
                            return false;
                        }
                    }
                }
            }
            Object[] array3 = new Object[length - 1];
            System.arraycopy(array2, 0, array3, 0, index);
            System.arraycopy(array2, index + 1, array3, index, length - index - 1);
            this.setArray(array3);
            return true;
        }
        finally {
            lock.unlock();
        }
    }
    
    void removeRange(int n, int n2) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = this.getArray();
            int length = array.length;
            if (n < 0 || n2 > length || n2 < n) {
                throw new IndexOutOfBoundsException();
            }
            int newLength = length - (n2 - n);
            int n3 = length - n2;
            if (n3 == 0) {
                this.setArray(Arrays.copyOf(array, newLength));
            }
            else {
                Object[] array2 = new Object[newLength];
                System.arraycopy(array, 0, array2, 0, n);
                System.arraycopy(array, n2, array2, n, n3);
                this.setArray(array2);
            }
        }
        finally {
            lock.unlock();
        }
    }
    
    public boolean addIfAbsent(E e) {
        Object[] array = this.getArray();
        return indexOf(e, array, 0, array.length) < 0 && this.addIfAbsent(e, array);
    }
    
    private boolean addIfAbsent(E e, Object[] array) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array2 = this.getArray();
            int length = array2.length;
            if (array != array2) {
                int min = Math.min(array.length, length);
                int i = 0;
                while (i < min) {
                    if (array2[i] != array[i] && eq(e, array2[i])) {
                        return false;
                    }
                    ++i;
                }
                if (indexOf(e, array2, min, length) >= 0) {
                    return false;
                }
            }
            Object[] copy = Arrays.copyOf(array2, length + 1);
            copy[length] = e;
            this.setArray(copy);
            return true;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean containsAll(Collection<?> collection) {
        Object[] array = this.getArray();
        int length = array.length;
        Iterator<Object> iterator = collection.iterator();
        while (iterator.hasNext()) {
            if (indexOf(iterator.next(), array, 0, length) < 0) {
                return false;
            }
            continue;
        }
        return true;
    }
    
    @Override
    public boolean removeAll(Collection<?> collection) {
        if (collection == null) {
            throw new NullPointerException();
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = this.getArray();
            int length = array.length;
            if (length != 0) {
                int newLength = 0;
                Object[] original = new Object[length];
                for (Object o : array) {
                    if (!collection.contains(o)) {
                        original[newLength++] = o;
                    }
                }
                if (newLength != length) {
                    this.setArray(Arrays.copyOf(original, newLength));
                    return true;
                }
            }
            return false;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean retainAll(Collection<?> collection) {
        if (collection == null) {
            throw new NullPointerException();
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = this.getArray();
            int length = array.length;
            if (length != 0) {
                int newLength = 0;
                Object[] original = new Object[length];
                for (Object o : array) {
                    if (collection.contains(o)) {
                        original[newLength++] = o;
                    }
                }
                if (newLength != length) {
                    this.setArray(Arrays.copyOf(original, newLength));
                    return true;
                }
            }
            return false;
        }
        finally {
            lock.unlock();
        }
    }
    
    public int addAllAbsent(Collection<? extends E> collection) {
        Object[] array = collection.toArray();
        if (collection.getClass() != ArrayList.class) {
            array = (Object[])array.clone();
        }
        if (array.length == 0) {
            return 0;
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array2 = this.getArray();
            int length = array2.length;
            int n = 0;
            for (int i = 0; i < array.length; ++i) {
                Object o = array[i];
                if (indexOf(o, array2, 0, length) < 0 && indexOf(o, array, 0, n) < 0) {
                    array[n++] = o;
                }
            }
            if (n > 0) {
                Object[] copy = Arrays.copyOf(array2, length + n);
                System.arraycopy(array, 0, copy, length, n);
                this.setArray(copy);
            }
            return n;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public void clear() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            this.setArray(new Object[0]);
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean addAll(Collection<? extends E> collection) {
        Object[] array = (collection.getClass() == CopyOnWriteArrayList.class) ? ((CopyOnWriteArrayList)collection).getArray() : collection.toArray();
        if (array.length == 0) {
            return false;
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array2 = this.getArray();
            int length = array2.length;
            if (length == 0 && (collection.getClass() == CopyOnWriteArrayList.class || collection.getClass() == ArrayList.class)) {
                this.setArray(array);
            }
            else {
                Object[] copy = Arrays.copyOf(array2, length + array.length);
                System.arraycopy(array, 0, copy, length, array.length);
                this.setArray(copy);
            }
            return true;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean addAll(int i, Collection<? extends E> collection) {
        Object[] array = collection.toArray();
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array2 = this.getArray();
            int length = array2.length;
            if (i > length || i < 0) {
                throw new IndexOutOfBoundsException("Index: " + i + ", Size: " + length);
            }
            if (array.length == 0) {
                return false;
            }
            int n = length - i;
            Object[] copy;
            if (n == 0) {
                copy = Arrays.copyOf(array2, length + array.length);
            }
            else {
                copy = new Object[length + array.length];
                System.arraycopy(array2, 0, copy, 0, i);
                System.arraycopy(array2, i, copy, i + array.length, n);
            }
            System.arraycopy(array, 0, copy, i, array.length);
            this.setArray(copy);
            return true;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public void forEach(Consumer<? super E> consumer) {
        if (consumer == null) {
            throw new NullPointerException();
        }
        Object[] array = this.getArray();
        for (int length = array.length, i = 0; i < length; ++i) {
            consumer.accept(array[i]);
        }
        return;
    }
    
    @Override
    public boolean removeIf(Predicate<? super E> predicate) {
        if (predicate == null) {
            throw new NullPointerException();
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = this.getArray();
            int length = array.length;
            if (length != 0) {
                int newLength = 0;
                Object[] original = new Object[length];
                for (Object o : array) {
                    if (!predicate.test(o)) {
                        original[newLength++] = o;
                    }
                }
                if (newLength != length) {
                    this.setArray(Arrays.copyOf(original, newLength));
                    return true;
                }
            }
            return false;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public void replaceAll(UnaryOperator<E> unaryOperator) {
        if (unaryOperator == null) {
            throw new NullPointerException();
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = this.getArray();
            int length = array.length;
            Object[] copy = Arrays.copyOf(array, length);
            for (int i = 0; i < length; ++i) {
                copy[i] = unaryOperator.apply(array[i]);
            }
            this.setArray(copy);
        }
        finally {
            lock.unlock();
        }
        return;
    }
    
    @Override
    public void sort(Comparator<? super E> c) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Object[] array = this.getArray();
            Object[] copy = Arrays.copyOf(array, array.length);
            Arrays.sort((Object[])copy, c);
            this.setArray(copy);
        }
        finally {
            lock.unlock();
        }
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        objectOutputStream.defaultWriteObject();
        Object[] array = this.getArray();
        objectOutputStream.writeInt(array.length);
        for (int length = array.length, i = 0; i < length; ++i) {
            objectOutputStream.writeObject(array[i]);
        }
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        objectInputStream.defaultReadObject();
        this.resetLock();
        int int1 = objectInputStream.readInt();
        SharedSecrets.getJavaOISAccess().checkArray(objectInputStream, Object[].class, int1);
        Object[] array = new Object[int1];
        for (int i = 0; i < int1; ++i) {
            array[i] = objectInputStream.readObject();
        }
        this.setArray(array);
    }
    
    @Override
    public String toString() {
        return Arrays.toString(this.getArray());
    }
    
    @Override
    public boolean equals(Object o) {
        if (o == this) {
            return true;
        }
        if (!(o instanceof List)) {
            return false;
        }
        Iterator<Object> iterator = ((List)(List)o).iterator();
        Object[] array = this.getArray();
        int length = array.length;
        int i = 0;
        while (i < length) {
            if (!iterator.hasNext() || !eq(array[i], iterator.next())) {
                return false;
            }
            ++i;
        }
        return !iterator.hasNext();
    }
    
    @Override
    public int hashCode() {
        int n = 1;
        for (Object o : this.getArray()) {
            n = 31 * n + ((o == null) ? 0 : o.hashCode());
        }
        return n;
    }
    
    @Override
    public Iterator<E> iterator() {
        return new CopyOnWriteArrayList.COWIterator(this.getArray(), 0);
    }
    
    @Override
    public ListIterator<E> listIterator() {
        return new CopyOnWriteArrayList.COWIterator(this.getArray(), 0);
    }
    
    @Override
    public ListIterator<E> listIterator(int i) {
        Object[] array = this.getArray();
        int length = array.length;
        if (i < 0 || i > length) {
            throw new IndexOutOfBoundsException("Index: " + i);
        }
        return new CopyOnWriteArrayList.COWIterator(array, i);
    }
    
    @Override
    public Spliterator<E> spliterator() {
        return Spliterators.spliterator(this.getArray(), 1040);
    }
    
    @Override
    public List<E> subList(int n, int n2) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            int length = this.getArray().length;
            if (n < 0 || n2 > length || n > n2) {
                throw new IndexOutOfBoundsException();
            }
            return new CopyOnWriteArrayList.COWSubList(this, n, n2);
        }
        finally {
            lock.unlock();
        }
    }
    
    private void resetLock() {
        CopyOnWriteArrayList.UNSAFE.putObjectVolatile(this, CopyOnWriteArrayList.lockOffset, new ReentrantLock());
    }
    
    static {
        try {
            UNSAFE = Unsafe.getUnsafe();
            lockOffset = CopyOnWriteArrayList.UNSAFE.objectFieldOffset(CopyOnWriteArrayList.class.getDeclaredField("lock"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
    
    static final class COWIterator<E> implements ListIterator<E>
    {
        private final Object[] snapshot;
        private int cursor;
        
        private COWIterator(Object[] snapshot, int cursor) {
            this.cursor = cursor;
            this.snapshot = snapshot;
        }
        
        @Override
        public boolean hasNext() {
            return this.cursor < this.snapshot.length;
        }
        
        @Override
        public boolean hasPrevious() {
            return this.cursor > 0;
        }
        
        @Override
        public E next() {
            if (!this.hasNext()) {
                throw new NoSuchElementException();
            }
            return this.snapshot[this.cursor++];
        }
        
        @Override
        public E previous() {
            if (!this.hasPrevious()) {
                throw new NoSuchElementException();
            }
            Object[] snapshot = this.snapshot;
            int cursor = this.cursor - 1;
            this.cursor = cursor;
            return snapshot[cursor];
        }
        
        @Override
        public int nextIndex() {
            return this.cursor;
        }
        
        @Override
        public int previousIndex() {
            return this.cursor - 1;
        }
        
        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }
        
        @Override
        public void set(E e) {
            throw new UnsupportedOperationException();
        }
        
        @Override
        public void add(E e) {
            throw new UnsupportedOperationException();
        }
        
        @Override
        public void forEachRemaining(Consumer<? super E> obj) {
            Objects.requireNonNull(obj);
            Object[] snapshot = this.snapshot;
            int length = snapshot.length;
            for (int i = this.cursor; i < length; ++i) {
                obj.accept(snapshot[i]);
            }
            this.cursor = length;
        }
    }
    
    private static class COWSubList<E> extends AbstractList<E> implements RandomAccess
    {
        private final CopyOnWriteArrayList<E> l;
        private final int offset;
        private int size;
        private Object[] expectedArray;
        
        COWSubList(CopyOnWriteArrayList<E> l, int offset, int n) {
            this.l = l;
            this.expectedArray = this.l.getArray();
            this.offset = offset;
            this.size = n - offset;
        }
        
        private void checkForComodification() {
            if (this.l.getArray() != this.expectedArray) {
                throw new ConcurrentModificationException();
            }
            return;
        }
        
        private void rangeCheck(int i) {
            if (i < 0 || i >= this.size) {
                throw new IndexOutOfBoundsException("Index: " + i + ",Size: " + this.size);
            }
            return;
        }
        
        @Override
        public E set(int n, E e) {
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                this.rangeCheck(n);
                this.checkForComodification();
                Object set = this.l.set(n + this.offset, e);
                this.expectedArray = this.l.getArray();
                return set;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public E get(int n) {
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                this.rangeCheck(n);
                this.checkForComodification();
                return this.l.get(n + this.offset);
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public int size() {
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                this.checkForComodification();
                return this.size;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public void add(int n, E e) {
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                this.checkForComodification();
                if (n < 0 || n > this.size) {
                    throw new IndexOutOfBoundsException();
                }
                this.l.add(n + this.offset, e);
                this.expectedArray = this.l.getArray();
                ++this.size;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public void clear() {
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                this.checkForComodification();
                this.l.removeRange(this.offset, this.offset + this.size);
                this.expectedArray = this.l.getArray();
                this.size = 0;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public E remove(int n) {
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                this.rangeCheck(n);
                this.checkForComodification();
                Object remove = this.l.remove(n + this.offset);
                this.expectedArray = this.l.getArray();
                --this.size;
                return remove;
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public boolean remove(Object o) {
            int index = this.indexOf(o);
            if (index == -1) {
                return false;
            }
            this.remove(index);
            return true;
        }
        
        @Override
        public Iterator<E> iterator() {
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                this.checkForComodification();
                return new CopyOnWriteArrayList.COWSubListIterator(this.l, 0, this.offset, this.size);
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public ListIterator<E> listIterator(int i) {
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                this.checkForComodification();
                if (i < 0 || i > this.size) {
                    throw new IndexOutOfBoundsException("Index: " + i + ", Size: " + this.size);
                }
                return new CopyOnWriteArrayList.COWSubListIterator(this.l, i, this.offset, this.size);
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public List<E> subList(int n, int n2) {
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                this.checkForComodification();
                if (n < 0 || n2 > this.size || n > n2) {
                    throw new IndexOutOfBoundsException();
                }
                return new CopyOnWriteArrayList.COWSubList(this.l, n + this.offset, n2 + this.offset);
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public void forEach(Consumer<? super E> consumer) {
            if (consumer == null) {
                throw new NullPointerException();
            }
            int i = this.offset;
            int n = this.offset + this.size;
            Object[] expectedArray = this.expectedArray;
            if (this.l.getArray() != expectedArray) {
                throw new ConcurrentModificationException();
            }
            if (i < 0 || n > expectedArray.length) {
                throw new IndexOutOfBoundsException();
            }
            for (; i < n; ++i) {
                consumer.accept(expectedArray[i]);
            }
            return;
        }
        
        @Override
        public void replaceAll(UnaryOperator<E> unaryOperator) {
            if (unaryOperator == null) {
                throw new NullPointerException();
            }
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                int i = this.offset;
                int n = this.offset + this.size;
                Object[] expectedArray = this.expectedArray;
                if (this.l.getArray() != expectedArray) {
                    throw new ConcurrentModificationException();
                }
                int length = expectedArray.length;
                if (i < 0 || n > length) {
                    throw new IndexOutOfBoundsException();
                }
                Object[] copy = Arrays.copyOf(expectedArray, length);
                for (; i < n; ++i) {
                    copy[i] = unaryOperator.apply(expectedArray[i]);
                }
                this.l.setArray(this.expectedArray = copy);
            }
            finally {
                lock.unlock();
            }
            return;
        }
        
        @Override
        public void sort(Comparator<? super E> c) {
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                int offset = this.offset;
                int toIndex = this.offset + this.size;
                Object[] expectedArray = this.expectedArray;
                if (this.l.getArray() != expectedArray) {
                    throw new ConcurrentModificationException();
                }
                int length = expectedArray.length;
                if (offset < 0 || toIndex > length) {
                    throw new IndexOutOfBoundsException();
                }
                Object[] copy = Arrays.copyOf(expectedArray, length);
                Arrays.sort((Object[])copy, offset, toIndex, c);
                this.l.setArray(this.expectedArray = copy);
            }
            finally {
                lock.unlock();
            }
        }
        
        @Override
        public boolean removeAll(Collection<?> collection) {
            if (collection == null) {
                throw new NullPointerException();
            }
            boolean n = false;
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                int size = this.size;
                if (size > 0) {
                    int i = this.offset;
                    int n2 = this.offset + size;
                    Object[] expectedArray = this.expectedArray;
                    if (this.l.getArray() != expectedArray) {
                        throw new ConcurrentModificationException();
                    }
                    int length = expectedArray.length;
                    if (i < 0 || n2 > length) {
                        throw new IndexOutOfBoundsException();
                    }
                    int size2 = 0;
                    Object[] array = new Object[size];
                    for (; i < n2; ++i) {
                        Object o = expectedArray[i];
                        if (!collection.contains(o)) {
                            array[size2++] = o;
                        }
                    }
                    if (size2 != size) {
                        Object[] expectedArray2 = new Object[length - size + size2];
                        System.arraycopy(expectedArray, 0, expectedArray2, 0, i);
                        System.arraycopy(array, 0, expectedArray2, i, size2);
                        System.arraycopy(expectedArray, n2, expectedArray2, i + size2, length - n2);
                        this.size = size2;
                        n = true;
                        this.l.setArray(this.expectedArray = expectedArray2);
                    }
                }
            }
            finally {
                lock.unlock();
            }
            return n;
        }
        
        @Override
        public boolean retainAll(Collection<?> collection) {
            if (collection == null) {
                throw new NullPointerException();
            }
            boolean n = false;
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                int size = this.size;
                if (size > 0) {
                    int i = this.offset;
                    int n2 = this.offset + size;
                    Object[] expectedArray = this.expectedArray;
                    if (this.l.getArray() != expectedArray) {
                        throw new ConcurrentModificationException();
                    }
                    int length = expectedArray.length;
                    if (i < 0 || n2 > length) {
                        throw new IndexOutOfBoundsException();
                    }
                    int size2 = 0;
                    Object[] array = new Object[size];
                    for (; i < n2; ++i) {
                        Object o = expectedArray[i];
                        if (collection.contains(o)) {
                            array[size2++] = o;
                        }
                    }
                    if (size2 != size) {
                        Object[] expectedArray2 = new Object[length - size + size2];
                        System.arraycopy(expectedArray, 0, expectedArray2, 0, i);
                        System.arraycopy(array, 0, expectedArray2, i, size2);
                        System.arraycopy(expectedArray, n2, expectedArray2, i + size2, length - n2);
                        this.size = size2;
                        n = true;
                        this.l.setArray(this.expectedArray = expectedArray2);
                    }
                }
            }
            finally {
                lock.unlock();
            }
            return n;
        }
        
        @Override
        public boolean removeIf(Predicate<? super E> predicate) {
            if (predicate == null) {
                throw new NullPointerException();
            }
            boolean n = false;
            ReentrantLock lock = this.l.lock;
            lock.lock();
            try {
                int size = this.size;
                if (size > 0) {
                    int i = this.offset;
                    int n2 = this.offset + size;
                    Object[] expectedArray = this.expectedArray;
                    if (this.l.getArray() != expectedArray) {
                        throw new ConcurrentModificationException();
                    }
                    int length = expectedArray.length;
                    if (i < 0 || n2 > length) {
                        throw new IndexOutOfBoundsException();
                    }
                    int size2 = 0;
                    Object[] array = new Object[size];
                    for (; i < n2; ++i) {
                        Object o = expectedArray[i];
                        if (!predicate.test(o)) {
                            array[size2++] = o;
                        }
                    }
                    if (size2 != size) {
                        Object[] expectedArray2 = new Object[length - size + size2];
                        System.arraycopy(expectedArray, 0, expectedArray2, 0, i);
                        System.arraycopy(array, 0, expectedArray2, i, size2);
                        System.arraycopy(expectedArray, n2, expectedArray2, i + size2, length - n2);
                        this.size = size2;
                        n = true;
                        this.l.setArray(this.expectedArray = expectedArray2);
                    }
                }
            }
            finally {
                lock.unlock();
            }
            return n;
        }
        
        @Override
        public Spliterator<E> spliterator() {
            int offset = this.offset;
            int toIndex = this.offset + this.size;
            Object[] expectedArray = this.expectedArray;
            if (this.l.getArray() != expectedArray) {
                throw new ConcurrentModificationException();
            }
            if (offset < 0 || toIndex > expectedArray.length) {
                throw new IndexOutOfBoundsException();
            }
            return Spliterators.spliterator(expectedArray, offset, toIndex, 1040);
        }
    }
    
    private static class COWSubListIterator<E> implements ListIterator<E>
    {
        private final ListIterator<E> it;
        private final int offset;
        private final int size;
        
        COWSubListIterator(List<E> list, int n, int offset, int size) {
            this.offset = offset;
            this.size = size;
            this.it = list.listIterator(n + offset);
        }
        
        @Override
        public boolean hasNext() {
            return this.nextIndex() < this.size;
        }
        
        @Override
        public E next() {
            if (this.hasNext()) {
                return this.it.next();
            }
            throw new NoSuchElementException();
        }
        
        @Override
        public boolean hasPrevious() {
            return this.previousIndex() >= 0;
        }
        
        @Override
        public E previous() {
            if (this.hasPrevious()) {
                return this.it.previous();
            }
            throw new NoSuchElementException();
        }
        
        @Override
        public int nextIndex() {
            return this.it.nextIndex() - this.offset;
        }
        
        @Override
        public int previousIndex() {
            return this.it.previousIndex() - this.offset;
        }
        
        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }
        
        @Override
        public void set(E e) {
            throw new UnsupportedOperationException();
        }
        
        @Override
        public void add(E e) {
            throw new UnsupportedOperationException();
        }
        
        @Override
        public void forEachRemaining(Consumer<? super E> obj) {
            Objects.requireNonNull(obj);
            int size = this.size;
            ListIterator<Object> it = this.it;
            while (this.nextIndex() < size) {
                obj.accept(it.next());
            }
        }
    }
}
